"""Exceptions module."""

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional, Union

from fred_types import ContactId, DomainId, KeysetId, NssetId, RegistrarId


class RegistryException(Exception):
    """Base exception for all registry exceptions."""


class HistoryChangeProhibited(RegistryException):
    """History change is prohibited."""


class InvalidData(RegistryException, ValueError):
    """Invalid data.

    This exception is DEPRECATED and will be eventually replaced by InvalidValue.
    """

    def __init__(self, fields: Sequence[str]) -> None:  # pragma: no cover
        super().__init__(fields)
        self.fields = fields


class InvalidHistoryInterval(RegistryException):
    """Invalid history interval."""


class InvalidSnapshot(RegistryException):
    """Invalid snapshot."""


@dataclass(init=True, eq=False)
class InvalidValue(RegistryException, ValueError):
    """Invalid value."""

    field: Sequence[Union[str, int]]
    detail: str = ""


class ObjectDoesNotExist(RegistryException):
    """Object does not exist.

    Base class for individual exceptions.
    """


class BillingError(RegistryException, RuntimeError):
    """Billing error (insufficient credit)."""


class PermissionDenied(RegistryException, RuntimeError):
    """Unauthorized operation attempt."""


@dataclass(init=True, eq=False)
class ContactDoesNotExist(ObjectDoesNotExist):
    """Contact does not exist."""

    id: Optional[ContactId] = None


@dataclass(init=True, eq=False)
class ContactBlocked(RegistryException):
    """Contact linking is prohibited."""

    id: ContactId


class ContactAlreadyExists(RegistryException):
    """Contact already exists."""


class ContactHandleInProtectionPeriod(RegistryException):
    """Contact with the same handle was recently deleted."""


@dataclass(init=True, eq=False)
class DomainDoesNotExist(ObjectDoesNotExist):
    """Domain does not exist."""

    id: Optional[DomainId] = None


@dataclass(init=True, eq=False)
class DomainBlocked(RegistryException):
    """Domain is blocked."""

    id: DomainId


@dataclass(init=True, eq=False)
class InvalidDomainRegistrationPeriod(RegistryException):
    """Invalid domain registration period."""

    too_short: bool = False
    too_long: bool = False
    policy_violation: bool = False


@dataclass(init=True, eq=False)
class InvalidFqdn(RegistryException):
    """Invalid FQDN."""

    invalid_format: bool = False
    unmanaged_zone: bool = False
    registered: bool = False
    auctioned: bool = False
    blacklisted: bool = False


@dataclass(init=True, eq=False)
class InvalidSequence(RegistryException):
    """Invalid sequence."""

    too_few_items: bool = False
    too_many_items: bool = False
    non_unique_items: bool = False


class ZoneDoesNotExist(ObjectDoesNotExist):
    """Zone does not exist."""


@dataclass(init=True, eq=False)
class KeysetDoesNotExist(ObjectDoesNotExist):
    """Keyset does not exist."""

    id: Optional[KeysetId] = None


@dataclass(init=True, eq=False)
class KeysetBlocked(RegistryException):
    """Keyset linking is prohibited."""

    id: KeysetId


@dataclass(init=True, eq=False)
class NssetDoesNotExist(ObjectDoesNotExist):
    """Nsset does not exist."""

    id: Optional[NssetId] = None


@dataclass(init=True, eq=False)
class RegistrarDoesNotExist(ObjectDoesNotExist):
    """Registrar does not exist."""

    id: Optional[RegistrarId] = None


class GroupDoesNotExist(ObjectDoesNotExist):
    """Registrar group does not exist."""


class RegistrarCertificationDoesNotExist(ObjectDoesNotExist):
    """Registrar certification does not exist."""


class RegistrarAlreadyExists(RegistryException):
    """Registrar already exists."""


class EppCredentialsDoesNotExist(RegistryException):
    """Registrar EPP credentials does not exist."""


class RepresentativeDoesNotExist(ObjectDoesNotExist):
    """Contact representative does not exist."""


class RepresentativeAlreadyExists(RegistryException):
    """Contact representative already exists."""


class BlockDoesNotExist(ObjectDoesNotExist):
    """Block does not exist."""


class BlockIdAlreadyExists(RegistryException):
    """Block id already exists."""


class NotCurrentVersion(RegistryException):
    """Contact history identification not valid."""


class ExtraArgumentWarning(UserWarning):
    """Extra argument was passed to a model."""


@dataclass(init=True, eq=False)
class StateFlagDoesNotExist(RegistryException):
    """State flag does not exist."""

    flag: str = ""
